//============================================================================
// Name        : Entry.cpp
// Author      : Zdeněk Kalčík
// Copyright   : @2009@
// Project 		 : BDS
//============================================================================

#include "Entry.hpp"
#include "TextUtil.hpp"

Entry::Entry() {

}

Entry::Entry(string outputName) {
	output = outputName;
	singularity = false;
	canInit = false;
}

Entry::~Entry() {
}

void Entry::addInput(string input) {
	inputs.push_back(input);
}

void Entry::addSetting(string setting) {
	size_t found = setting.find_first_of(' ');
	string set = setting.substr(0, found);
	if (set.length() != inputs.size()) {
		cerr << "bad variables count in definition of " << output << endl;
		exit(1);
	}
	string output = setting.substr(found + 1);
	output = trimS(output);
	if (output.data()[0] == '0') {
		singularity = true;
	}
	settings.push_back(set);
}

string Entry::getOutput() {
	return output;
}

void Entry::toString() {
	cout << "Entry : " << output << endl << "inputs: ";
	for (unsigned int i = 0; i < inputs.size(); i++) {
		cout << inputs[i] << " ";
	}
	cout << endl << "settings: ";
	for (unsigned int i = 0; i < settings.size(); i++) {
		cout << settings[i] << endl;
	}
	cout << "vystup: ";
	if (singularity) {
		cout << 0 << endl << endl;
	} else {
		cout << 1 << endl << endl;
	}
	cout.flush();
}

bool Entry::canInitialize(set<string>& inilializedNodes) {
	if (canInit) {
		return true;
	}
	for (unsigned int i = 0; i < inputs.size(); i++) {
		if (inilializedNodes.find(inputs[i]) == inilializedNodes.end()) {
			return false;
		}
	}
	canInit = true;
	return true;
}

int Entry::lineCount() {
	return settings.size();
}

string Entry::getLine(int number) {
	return settings[number];
}

string Entry::getInput(int number) {
	return inputs[number];
}

bool Entry::getSingularity() {
	return singularity;
}
