//============================================================================
// Name        : Manager.cpp
// Author      : Zdeněk Kalčík
// Copyright   : @2009@
// Project 		 : BDS
//============================================================================

#include "Manager.hpp"
#include "Decompositions.hpp"

set<int> cut2;

//key = index, value = cout leaf edges
map<int, int> leafEdges0;
map<int, int> leafEdges1;

//key = index, value 1 = Regular, 2 = Complemented, 3 = generalized x-dominator
map<string, int> gxd;
map<string, DdNode*> gxdNodes;
map<string, int>::iterator gxdIterator;

set<string> visitedNodes;

map<int, int>::iterator it;
DdNode *add;

void findLeafEges(Decomp* decom);
void analyzeLeaf(DdNode *node);
void findGXD(Decomp* decom);
void analyzeGXD(DdNode *node);
void createCut(Decomp *decomp, int cutLevel);

int maxAndCut;
int indexMaxAndCut;

int maxOrCut;
int indexMaxOrCut;

int decompChoosen;

vector<Decomp*> decompositions;
Decomp *decomp;

int chooseDecomposition();
int getGXDCount();
void searchIndexMaxAndCut();
void searchIndexMaxOrCut();

void doDecomposition(Decomp* decomposition, char reoType) {
	initReorderType(reoType);

	decompositions.clear();
	decompositions.push_back(decomposition);

	while (!decompositions.empty()) {
		decomp = decompositions.back();
		decompositions.pop_back();
		findLeafEges(decomp);
		findGXD(decomp);
		decompChoosen = chooseDecomposition();
		switch (decompChoosen) {
			case AND: {
				//cout << "choosing AND decomposition, max: " << maxAndCut << " for index: " << indexMaxAndCut << endl;
				createCut(decomp, indexMaxAndCut);
				doAND(decomp);
				decompositions.push_back(decomp -> L);
				decompositions.push_back(decomp -> R);
				break;
			}
			case OR: {
				//cout << "choosing OR decomposition, max: " << maxOrCut << " for index: " << indexMaxOrCut << endl;
				createCut(decomp, indexMaxOrCut);
				doOR(decomp);
				decompositions.push_back(decomp -> L); //H
				decompositions.push_back(decomp -> R); //G
				break;
			}
			case XNOR: {
				//cout << "choosing XNOR" << endl;
				for (gxdIterator = gxd.begin(); gxdIterator != gxd.end(); gxdIterator++) {
					if ((*gxdIterator).second == 3) {
						initGNX(gxdNodes[(*gxdIterator).first]);
						//break;
					}
				}
				doXNOR(decomp);
				decompositions.push_back(decomp -> L);
				decompositions.push_back(decomp -> R);
				break;
			}
			default:
				//it's not possible to do another decomposition
				break;
		}
	}
}

/**
 * Return best decomposition type. Must be call after findLeafEges(Decomp*) and findGXD(Decomp*)
 *
 * Side Effects: It sets maxAndCut, indexMaxAndCut, maxOrCut, indexMaxOrCut
 *
 */
int chooseDecomposition() {
	maxAndCut = 0;
	indexMaxAndCut = 0;

	maxOrCut = 0;
	indexMaxOrCut = 0;

	int posibbleAndCutCount = leafEdges0.size();
	searchIndexMaxAndCut();
	int posibbleOrCutCount = leafEdges1.size();
	searchIndexMaxOrCut();
	int GXDcount = getGXDCount();

	//cout << "\n" << endl;
	//cout << "AND cut count: " << posibbleAndCutCount << "  max: " << maxAndCut << " (" <<indexMaxAndCut << ")" << endl;
	//cout << "OR cut count: " << posibbleOrCutCount << "  max: " << maxOrCut << " (" <<indexMaxOrCut << ")" << endl;
	//cout << "GXDs count: " << GXDcount << endl;
	//cout << "\n" << endl;

	if (maxAndCut == 0 && maxOrCut == 0 && GXDcount == 0) {
		//anything else to do
		return 0;
	}

	//simple heuristic
	//if (true || (posibbleAndCutCount + posibbleOrCutCount) > GXDcount) {
	if ((posibbleAndCutCount + posibbleOrCutCount) > 0) {
		if (maxAndCut > maxOrCut) {
			return AND;
		} else {
			return OR;
		}
	} else {
		return XNOR;
	}
}

/**
 * Search max count of cuts for AND decompsition.
 *
 * Side Effects: It sets maxAndCut, indexMaxAndCut
 *
 */
void searchIndexMaxAndCut() {
	for (it = leafEdges0.begin(); it != leafEdges0.end(); it++) {
		//cout << "na indexu: "  << (*it).first << "  and: " << (*it).second;
		if ((*it).second > maxAndCut) {
			maxAndCut = (*it).second;
			indexMaxAndCut = (*it).first;
		}
	}
}

/**
 * Search max count of cuts for OR decompsition.
 *
 * Side Effects: It sets maxOrCut, indexMaxOrCut
 *
 */
void searchIndexMaxOrCut() {
	for (it = leafEdges1.begin(); it != leafEdges1.end(); it++) {
		if ((*it).second > maxOrCut) {
			maxOrCut = (*it).second;
			indexMaxOrCut = (*it).first;
		}
	}
}

/**
 * Return count of generalized x-dominators.
 */
int getGXDCount() {
	int GXDcount = 0;
	map<string, int>::iterator ite;
	for (ite = gxd.begin(); ite != gxd.end(); ite++) {
		//cout << (*ite).first << " GXD=> " << (*ite).second << endl;
		if ((*ite).second == 3) {
			GXDcount++;
		}
	}
	return GXDcount;
}

/**
 * Create set of variables above the cut. It's not affected by reordering. It fills set with variable's index.
 * cutLevel is index of variable for horizontal cut.
 *
 * Side Effects: It sets cut2.
 *
 */
void createCut(Decomp *decomp, int cutLevel) {
	cut2.clear();
	for (int i = 0; i < Cudd_ReadSize(decomp -> manag); i++) {
		//cout << "on index " << i << " there is variable  " << Cudd_ReadInvPerm(decomp -> manag, i) << endl;
		cut2.insert(Cudd_ReadInvPerm(decomp -> manag, i));
		if (Cudd_ReadInvPerm(decomp -> manag, i) == cutLevel) {
			break;
		}
	}
	initCut(cut2);
}

/**
 * Find all 0 and 1 leaf edges.
 *
 * Side Effects: It sets leafEdges0 and leafEdges1.
 *
 */
void findLeafEges(Decomp* decomp) {
	leafEdges0.clear();
	leafEdges1.clear();
	visitedNodes.clear();
	add = Cudd_BddToAdd(decomp -> manag, decomp -> function);
	Cudd_Ref(add);
	analyzeLeaf(add);
	Cudd_RecursiveDeref(decomp -> manag, add);
}

/**
 * Find all generalized x-dominators..
 *
 * Side Effects: It sets gxd.
 *
 */
void findGXD(Decomp* decomp) {
	visitedNodes.clear();
	gxd.clear();
	gxdNodes.clear();
	analyzeGXD(decomp -> function);
}

void addR(string id) {
	if (gxd.find(id) == gxd.end()) {
		gxd[id] = 1;
	} else {
		if (gxd[id] == 2) {
			gxd[id] = 3;
		}
	}
}

void addC(string id, DdNode *node) {
	gxdNodes[id] = node;
	if (gxd.find(id) == gxd.end()) {
		gxd[id] = 2;
	} else {
		if (gxd[id] == 1) {
			gxd[id] = 3;
		}
	}
}

void analyzeGXD(DdNode *node) {
	if (!Cudd_IsConstant(Cudd_T(node))) {

		addR(getNodeIndex(Cudd_T(node)));

		if (visitedNodes.find(getNodeIndex(Cudd_T(node))) == visitedNodes.end()) {
			analyzeGXD(Cudd_T(node));
		}
	}

	if (!Cudd_IsConstant(Cudd_E(node))) {

		if (Cudd_IsComplement(Cudd_E(node))) {
			addC(getNodeIndex(Cudd_Regular(Cudd_E(node))), Cudd_Regular(Cudd_E(node)));
			if (visitedNodes.find(getNodeIndex(Cudd_Regular(Cudd_E(node)))) == visitedNodes.end()) {
				analyzeGXD(Cudd_Regular(Cudd_E(node)));
			}
		} else {
			addR(getNodeIndex(Cudd_E(node)));
			if (visitedNodes.find(getNodeIndex(Cudd_E(node))) == visitedNodes.end()) {
				analyzeGXD(Cudd_E(node));
			}
		}
	}

	visitedNodes.insert(getNodeIndex(node));
}

void add0(int index) {
	if (leafEdges0.find(index) == leafEdges0.end()) {
		leafEdges0[index] = 1;
	} else {
		leafEdges0[index]++;
	}
}

void add1(int index) {
	if (leafEdges1.find(index) == leafEdges1.end()) {
		leafEdges1[index] = 1;
	} else {
		leafEdges1[index]++;
	}
}

void analyzeLeaf(DdNode *node) {

	if (Cudd_IsConstant(Cudd_T(node))) {
		//if else-edge is constant too, then this cut is not valid
		if (!Cudd_IsConstant(Cudd_E(node))) {
			if (Cudd_V(Cudd_T(node)) == 1) {
				//add 1-leaf edge for index of node
				add1(node -> index);
			} else {
				//add 0-leaf edge for index of node
				add0(node -> index);
			}
		}
	} else {
		if (visitedNodes.find(getNodeIndex(Cudd_T(node))) == visitedNodes.end()) {
			analyzeLeaf(Cudd_T(node));
		}
	}

	if (Cudd_IsConstant(Cudd_E(node))) {
		//if then-edge is constant too, then this cut is not valid
		if (!Cudd_IsConstant(Cudd_T(node))) {
			if (Cudd_V(Cudd_E(node)) == 1) {
				//add 1-leaf edge for index of node
				add1(node -> index);
			} else {
				//add 0-leaf edge for index of node
				add0(node -> index);
			}
		}
	} else {
		if (visitedNodes.find(getNodeIndex(Cudd_E(node))) == visitedNodes.end()) {
			analyzeLeaf(Cudd_E(node));
		}
	}

	visitedNodes.insert(getNodeIndex(node));
}
